from .aerodynamics import *
from ..geometry import *


class Casvlm1(AeroProblem):
    # Usage:
    #   # Set up attrib_name problem using the syntax in the AeroProblem constructor (e.g. "Casvlm1(airplane = attrib_name, op_point = op)" for some Airplane attrib_name and OperatingPoint op)
    #   # Call the run() method on the vlm3 object to run the problem.
    #   # Access results in the command line, or through properties of the Casvlm1 class.
    #   #   # In attrib_name future update, this will be done through attrib_name standardized AeroData class.

    def __init__(self,
                 airplane,  # type: Airplane
                 op_point,  # type: op_point
                 opti,  # type: cas.Opti
                 run_setup=True,
                 ):
        super().__init__(airplane, op_point)
        self.opti = opti

        if run_setup:
            self.setup()

    def setup(self, verbose=True):
        # Runs attrib_name point analysis at the specified op-point.
        self.verbose = verbose

        if self.verbose:
            print("Setting up casVLM1 calculation...")

        self.make_panels()
        self.setup_geometry()
        self.setup_operating_point()
        self.calculate_vortex_strengths()
        self.calculate_forces()

        if self.verbose:
            print("casVLM1 setup complete! Ready to pass into the solver...")

    def make_panels(self):
        # Creates self.panel_coordinates_structured_list and self.wing_mcl_normals.

        if self.verbose:
            print("Meshing...")

        front_left_vertices = []
        front_right_vertices = []
        back_left_vertices = []
        back_right_vertices = []
        is_trailing_edge = []
        normal_directions = []

        for wing_num in range(len(self.airplane.wings)):
            # Things we want for each wing (where M is the number of chordwise panels, N is the number of spanwise panels)
            # # panel_coordinates_structured_list: M+1 p N+1 p 3; corners of every panel.
            # # normals_structured_list: M p N p 3; normal direction of each panel

            # Get the wing
            wing = self.airplane.wings[wing_num]  # type: Wing

            # Define number of chordwise points
            n_chordwise_coordinates = wing.chordwise_panels + 1

            # Get the chordwise coordinates
            if wing.chordwise_spacing == 'uniform':
                nondim_chordwise_coordinates = np.linspace(0, 1, n_chordwise_coordinates)
            elif wing.chordwise_spacing == 'cosine':
                nondim_chordwise_coordinates = np_cosspace(0, 1, n_chordwise_coordinates)
            else:
                raise Exception("Bad init_val of wing.chordwise_spacing!")

            # -----------------------------------------------------
            ## Make the panels for each section.

            for section_num in range(len(wing.xsecs) - 1):
                # Define the relevant cross sections
                inner_xsec = wing.xsecs[section_num]  # type: WingXSec
                outer_xsec = wing.xsecs[section_num + 1]  # type: WingXSec

                # Find the corners
                inner_xsec_xyz_le = inner_xsec.xyz_le + wing.xyz_le
                inner_xsec_xyz_te = inner_xsec.xyz_te() + wing.xyz_le
                outer_xsec_xyz_le = outer_xsec.xyz_le + wing.xyz_le
                outer_xsec_xyz_te = outer_xsec.xyz_te() + wing.xyz_le

                # Define the airfoils at each cross section
                inner_airfoil = inner_xsec.airfoil.get_airfoil_with_control_surface(
                    deflection=inner_xsec.control_surface_deflection,
                    hinge_point=inner_xsec.control_surface_hinge_point
                )  # type: Airfoil
                outer_airfoil = outer_xsec.airfoil.get_airfoil_with_control_surface(
                    deflection=inner_xsec.control_surface_deflection,
                    hinge_point=inner_xsec.control_surface_hinge_point
                )  # type: Airfoil

                # Make the mean camber lines for each.
                inner_xsec_mcl_nondim = inner_airfoil.get_downsampled_mcl(nondim_chordwise_coordinates)
                outer_xsec_mcl_nondim = outer_airfoil.get_downsampled_mcl(nondim_chordwise_coordinates)

                # Find the tangent angles of the mean camber lines
                inner_xsec_mcl_angle = (
                        cas.atan2(cas.diff(inner_xsec_mcl_nondim[:, 1]), cas.diff(inner_xsec_mcl_nondim[:, 0])) - (
                        inner_xsec.twist * np.pi / 180)
                )  # in radians
                outer_xsec_mcl_angle = (
                        cas.atan2(cas.diff(outer_xsec_mcl_nondim[:, 1]), cas.diff(outer_xsec_mcl_nondim[:, 0])) - (
                        outer_xsec.twist * np.pi / 180)
                )  # in radians

                # Find the effective twist axis
                effective_twist_axis = outer_xsec_xyz_le - inner_xsec_xyz_le
                effective_twist_axis[0] = 0.

                # Define number of spanwise points
                n_spanwise_coordinates = inner_xsec.spanwise_panels + 1

                # Get the spanwise coordinates
                if inner_xsec.spanwise_spacing == 'uniform':
                    nondim_spanwise_coordinates = np.linspace(0, 1, n_spanwise_coordinates)
                elif inner_xsec.spanwise_spacing == 'cosine':
                    nondim_spanwise_coordinates = np_cosspace(0, 1, n_spanwise_coordinates)
                else:
                    raise Exception("Bad init_val of section.spanwise_spacing!")

                for chord_index in range(wing.chordwise_panels):
                    for span_index in range(inner_xsec.spanwise_panels):
                        nondim_chordwise_coordinate = nondim_chordwise_coordinates[chord_index]
                        nondim_spanwise_coordinate = nondim_spanwise_coordinates[span_index]
                        nondim_chordwise_coordinate_next = nondim_chordwise_coordinates[chord_index + 1]
                        nondim_spanwise_coordinate_next = nondim_spanwise_coordinates[span_index + 1]

                        # Calculate vertices
                        front_left_vertices.append(
                            (
                                    inner_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate) + inner_xsec_xyz_te * nondim_chordwise_coordinate
                            ) * (1 - nondim_spanwise_coordinate) +
                            (
                                    outer_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate) + outer_xsec_xyz_te * nondim_chordwise_coordinate
                            ) * nondim_spanwise_coordinate
                        )
                        front_right_vertices.append(
                            (
                                    inner_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate) + inner_xsec_xyz_te * nondim_chordwise_coordinate
                            ) * (1 - nondim_spanwise_coordinate_next) +
                            (
                                    outer_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate) + outer_xsec_xyz_te * nondim_chordwise_coordinate
                            ) * nondim_spanwise_coordinate_next
                        )
                        back_left_vertices.append(
                            (
                                    inner_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate_next) + inner_xsec_xyz_te * nondim_chordwise_coordinate_next
                            ) * (1 - nondim_spanwise_coordinate) +
                            (
                                    outer_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate_next) + outer_xsec_xyz_te * nondim_chordwise_coordinate_next
                            ) * nondim_spanwise_coordinate
                        )
                        back_right_vertices.append(
                            (
                                    inner_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate_next) + inner_xsec_xyz_te * nondim_chordwise_coordinate_next
                            ) * (1 - nondim_spanwise_coordinate_next) +
                            (
                                    outer_xsec_xyz_le * (
                                    1 - nondim_chordwise_coordinate_next) + outer_xsec_xyz_te * nondim_chordwise_coordinate_next
                            ) * nondim_spanwise_coordinate_next
                        )
                        is_trailing_edge.append(chord_index == wing.chordwise_panels - 1)

                        # Calculate normal
                        angle = (
                                inner_xsec_mcl_angle[chord_index] * (1 - nondim_spanwise_coordinate) +
                                inner_xsec_mcl_angle[chord_index] * nondim_spanwise_coordinate
                        )
                        rot = angle_axis_rotation_matrix(-angle - np.pi / 2, effective_twist_axis)
                        normal_directions.append(rot @ cas.vertcat(1, 0, 0))

                # Handle symmetry
                if wing.symmetric:
                    # Define the relevant cross sections
                    inner_xsec = wing.xsecs[section_num]  # type: WingXSec
                    outer_xsec = wing.xsecs[section_num + 1]  # type: WingXSec

                    # Define the airfoils at each cross section
                    if inner_xsec.control_surface_type == "symmetric":
                        inner_airfoil = inner_xsec.airfoil.get_airfoil_with_control_surface(
                            deflection=inner_xsec.control_surface_deflection,
                            hinge_point=inner_xsec.control_surface_hinge_point
                        )  # type: Airfoil
                        outer_airfoil = outer_xsec.airfoil.get_airfoil_with_control_surface(
                            deflection=inner_xsec.control_surface_deflection,
                            hinge_point=inner_xsec.control_surface_hinge_point
                        )  # type: Airfoil
                    elif inner_xsec.control_surface_type == "asymmetric":
                        inner_airfoil = inner_xsec.airfoil.get_airfoil_with_control_surface(
                            deflection=-inner_xsec.control_surface_deflection,
                            hinge_point=inner_xsec.control_surface_hinge_point
                        )  # type: Airfoil
                        outer_airfoil = outer_xsec.airfoil.get_airfoil_with_control_surface(
                            deflection=-inner_xsec.control_surface_deflection,
                            hinge_point=inner_xsec.control_surface_hinge_point
                        )  # type: Airfoil
                    else:
                        raise ValueError("Invalid input for control_surface_type!")

                    # Make the mean camber lines for each.
                    inner_xsec_mcl_nondim = inner_airfoil.get_downsampled_mcl(nondim_chordwise_coordinates)
                    outer_xsec_mcl_nondim = outer_airfoil.get_downsampled_mcl(nondim_chordwise_coordinates)

                    # Find the tangent angles of the mean camber lines
                    inner_xsec_mcl_angle = (
                            cas.atan2(cas.diff(inner_xsec_mcl_nondim[:, 1]),
                                      cas.diff(inner_xsec_mcl_nondim[:, 0])) - (
                                    inner_xsec.twist * np.pi / 180)
                    )  # in radians
                    outer_xsec_mcl_angle = (
                            cas.atan2(cas.diff(outer_xsec_mcl_nondim[:, 1]),
                                      cas.diff(outer_xsec_mcl_nondim[:, 0])) - (
                                    outer_xsec.twist * np.pi / 180)
                    )  # in radians

                    # Find the effective twist axis
                    effective_twist_axis = outer_xsec_xyz_le - inner_xsec_xyz_le
                    effective_twist_axis[0] = 0.

                    # Define number of spanwise points
                    n_spanwise_coordinates = inner_xsec.spanwise_panels + 1

                    # Get the spanwise coordinates
                    if inner_xsec.spanwise_spacing == 'uniform':
                        nondim_spanwise_coordinates = np.linspace(0, 1, n_spanwise_coordinates)
                    elif inner_xsec.spanwise_spacing == 'cosine':
                        nondim_spanwise_coordinates = np_cosspace(0, 1, n_spanwise_coordinates)
                    else:
                        raise Exception("Bad init_val of section.spanwise_spacing!")

                    for chord_index in range(wing.chordwise_panels):
                        for span_index in range(inner_xsec.spanwise_panels):
                            nondim_chordwise_coordinate = nondim_chordwise_coordinates[chord_index]
                            nondim_spanwise_coordinate = nondim_spanwise_coordinates[span_index]
                            nondim_chordwise_coordinate_next = nondim_chordwise_coordinates[chord_index + 1]
                            nondim_spanwise_coordinate_next = nondim_spanwise_coordinates[span_index + 1]

                            # Calculate vertices
                            front_right_vertices.append(reflect_over_XZ_plane(
                                (
                                        inner_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate) + inner_xsec_xyz_te * nondim_chordwise_coordinate
                                ) * (1 - nondim_spanwise_coordinate) +
                                (
                                        outer_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate) + outer_xsec_xyz_te * nondim_chordwise_coordinate
                                ) * nondim_spanwise_coordinate
                            ))
                            front_left_vertices.append(reflect_over_XZ_plane(
                                (
                                        inner_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate) + inner_xsec_xyz_te * nondim_chordwise_coordinate
                                ) * (1 - nondim_spanwise_coordinate_next) +
                                (
                                        outer_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate) + outer_xsec_xyz_te * nondim_chordwise_coordinate
                                ) * nondim_spanwise_coordinate_next
                            ))
                            back_right_vertices.append(reflect_over_XZ_plane(
                                (
                                        inner_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate_next) + inner_xsec_xyz_te * nondim_chordwise_coordinate_next
                                ) * (1 - nondim_spanwise_coordinate) +
                                (
                                        outer_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate_next) + outer_xsec_xyz_te * nondim_chordwise_coordinate_next
                                ) * nondim_spanwise_coordinate
                            ))
                            back_left_vertices.append(reflect_over_XZ_plane(
                                (
                                        inner_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate_next) + inner_xsec_xyz_te * nondim_chordwise_coordinate_next
                                ) * (1 - nondim_spanwise_coordinate_next) +
                                (
                                        outer_xsec_xyz_le * (
                                        1 - nondim_chordwise_coordinate_next) + outer_xsec_xyz_te * nondim_chordwise_coordinate_next
                                ) * nondim_spanwise_coordinate_next
                            ))
                            is_trailing_edge.append(chord_index == wing.chordwise_panels - 1)

                            # Calculate normal
                            angle = (
                                    inner_xsec_mcl_angle[chord_index] * (1 - nondim_spanwise_coordinate) +
                                    inner_xsec_mcl_angle[chord_index] * nondim_spanwise_coordinate
                            )
                            rot = angle_axis_rotation_matrix(-angle - np.pi / 2, effective_twist_axis)
                            normal_directions.append(reflect_over_XZ_plane(rot @ cas.vertcat(1, 0, 0)))

        # Concatenate things (DM)
        # self.front_left_vertices = cas.transpose(cas.horzcat(*front_left_vertices))
        # self.front_right_vertices = cas.transpose(cas.horzcat(*front_right_vertices))
        # self.back_left_vertices = cas.transpose(cas.horzcat(*back_left_vertices))
        # self.back_right_vertices = cas.transpose(cas.horzcat(*back_right_vertices))
        # self.normal_directions = cas.transpose(cas.horzcat(*normal_directions))
        # self.is_trailing_edge = is_trailing_edge

        # Concatenate things (MX)
        self.front_left_vertices = cas.MX(cas.transpose(cas.horzcat(*front_left_vertices)))
        self.front_right_vertices = cas.MX(cas.transpose(cas.horzcat(*front_right_vertices)))
        self.back_left_vertices = cas.MX(cas.transpose(cas.horzcat(*back_left_vertices)))
        self.back_right_vertices = cas.MX(cas.transpose(cas.horzcat(*back_right_vertices)))
        self.normal_directions = cas.MX(cas.transpose(cas.horzcat(*normal_directions)))
        self.is_trailing_edge = is_trailing_edge

        # Calculate areas
        diag1 = self.front_right_vertices - self.back_left_vertices
        diag2 = self.front_left_vertices - self.back_right_vertices
        cross = cas.cross(diag1, diag2)
        cross_norm = cas.sqrt(cross[:, 0] ** 2 + cross[:, 1] ** 2 + cross[:, 2] ** 2)
        self.areas = cross_norm / 2

        # Do the vortex math
        self.left_vortex_vertices = 0.75 * self.front_left_vertices + 0.25 * self.back_left_vertices
        self.right_vortex_vertices = 0.75 * self.front_right_vertices + 0.25 * self.back_right_vertices
        self.vortex_centers = (self.left_vortex_vertices + self.right_vortex_vertices) / 2
        self.vortex_bound_leg = (self.right_vortex_vertices - self.left_vortex_vertices)
        self.collocation_points = (
                0.5 * (
                0.25 * self.front_left_vertices + 0.75 * self.back_left_vertices
        ) +
                0.5 * (
                        0.25 * self.front_right_vertices + 0.75 * self.back_right_vertices
                )
        )

        # Do final processing for later use
        self.n_panels = self.collocation_points.shape[0]

        if self.verbose:
            print("Meshing complete!")

    def setup_geometry(self):
        # # Calculate AIC matrix
        # ----------------------
        if self.verbose:
            print("Calculating the collocation influence matrix...")
        self.Vij_collocations_x, self.Vij_collocations_y, self.Vij_collocations_z = self.calculate_Vij(
            self.collocation_points)

        # AIC = (Vij * normal vectors)
        self.AIC = (
                self.Vij_collocations_x * self.normal_directions[:, 0] +
                self.Vij_collocations_y * self.normal_directions[:, 1] +
                self.Vij_collocations_z * self.normal_directions[:, 2]
        )

        # # Calculate Vij at vortex centers for force calculation
        # -------------------------------------------------------
        if self.verbose:
            print("Calculating the vortex center influence matrix...")
        self.Vij_centers_x, self.Vij_centers_y, self.Vij_centers_z = self.calculate_Vij(self.vortex_centers)

    def setup_operating_point(self):
        if self.verbose:
            print("Calculating the freestream influence...")
        self.steady_freestream_velocity = self.op_point.compute_freestream_velocity_geometry_axes()  # Direction the wind is GOING TO, in geometry axes coordinates
        self.rotation_freestream_velocities = self.op_point.compute_rotation_velocity_geometry_axes(
            self.collocation_points)

        self.freestream_velocities = cas.transpose(self.steady_freestream_velocity + cas.transpose(
            self.rotation_freestream_velocities))  # Nx3, represents the freestream velocity at each panel collocation point (c)

        self.freestream_influences = (
                self.freestream_velocities[:, 0] * self.normal_directions[:, 0] +
                self.freestream_velocities[:, 1] * self.normal_directions[:, 1] +
                self.freestream_velocities[:, 2] * self.normal_directions[:, 2]
        )

    def calculate_vortex_strengths(self):
        # # Calculate Vortex Strengths
        # ----------------------------
        # Governing Equation: AIC @ Gamma + freestream_influence = 0
        if self.verbose:
            print("Calculating vortex strengths...")

        # Explicit solve
        self.vortex_strengths = cas.solve(self.AIC, -self.freestream_influences)

        # # Implicit solve
        # self.vortex_strengths = self.opti.variable(self.n_panels)
        # self.opti.set_initial(self.vortex_strengths, 1)
        # self.opti.subject_to([
        #     self.AIC @ self.vortex_strengths == -self.freestream_influences
        # ])

    def calculate_forces(self):
        # # Calculate Near-Field Forces and Moments
        # -----------------------------------------
        # Governing Equation: The force on attrib_name straight, small vortex filament is F = rho * V p l * gamma,
        # where rho is density, V is the velocity vector, p is the cross product operator,
        # l is the vector of the filament itself, and gamma is the circulation.

        if self.verbose:
            print("Calculating forces on each panel...")
        # Calculate Vi (local velocity at the ith vortex center point)
        Vi_x = self.Vij_centers_x @ self.vortex_strengths + self.freestream_velocities[:, 0]
        Vi_y = self.Vij_centers_y @ self.vortex_strengths + self.freestream_velocities[:, 1]
        Vi_z = self.Vij_centers_z @ self.vortex_strengths + self.freestream_velocities[:, 2]
        Vi = cas.horzcat(Vi_x, Vi_y, Vi_z)

        # Calculate forces_inviscid_geometry, the force on the ith panel. Note that this is in GEOMETRY AXES,
        # not WIND AXES or BODY AXES.
        density = self.op_point.density
        # Vi_cross_li = np.cross(Vi, self.vortex_bound_leg, axis=1)
        Vi_cross_li = cas.horzcat(
            Vi_y * self.vortex_bound_leg[:, 2] - Vi_z * self.vortex_bound_leg[:, 1],
            Vi_z * self.vortex_bound_leg[:, 0] - Vi_x * self.vortex_bound_leg[:, 2],
            Vi_x * self.vortex_bound_leg[:, 1] - Vi_y * self.vortex_bound_leg[:, 0],
        )
        # vortex_strengths_expanded = np.expand_dims(self.vortex_strengths, axis=1)
        self.Fi_geometry = density * Vi_cross_li * self.vortex_strengths

        # Calculate total forces and moments
        if self.verbose:
            print("Calculating total forces and moments...")
        self.Ftotal_geometry = cas.vertcat(
            cas.sum1(self.Fi_geometry[:, 0]),
            cas.sum1(self.Fi_geometry[:, 1]),
            cas.sum1(self.Fi_geometry[:, 2]),
        )  # Remember, this is in GEOMETRY AXES, not WIND AXES or BODY AXES.
        # if self.verbose: print("Total aerodynamic forces (geometry axes): ", self.force_total_inviscid_geometry)

        self.Ftotal_wind = cas.transpose(
            self.op_point.compute_rotation_matrix_wind_to_geometry()) @ self.Ftotal_geometry
        # if self.verbose: print("Total aerodynamic forces (wind axes):", self.force_total_inviscid_wind)

        self.Mi_geometry = cas.cross(
            cas.transpose(cas.transpose(self.vortex_centers) - self.airplane.xyz_ref),
            self.Fi_geometry
        )

        self.Mtotal_geometry = cas.vertcat(
            cas.sum1(self.Mi_geometry[:, 0]),
            cas.sum1(self.Mi_geometry[:, 1]),
            cas.sum1(self.Mi_geometry[:, 2]),
        )

        self.Mtotal_wind = cas.transpose(
            self.op_point.compute_rotation_matrix_wind_to_geometry()) @ self.Mtotal_geometry

        # Calculate nondimensional forces
        q = self.op_point.dynamic_pressure()
        s_ref = self.airplane.s_ref
        b_ref = self.airplane.b_ref
        c_ref = self.airplane.c_ref
        self.CL = -self.Ftotal_wind[2] / q / s_ref
        self.CDi = -self.Ftotal_wind[0] / q / s_ref
        self.CY = self.Ftotal_wind[1] / q / s_ref
        self.Cl = self.Mtotal_wind[0] / q / s_ref / b_ref
        self.Cm = self.Mtotal_wind[1] / q / s_ref / c_ref
        self.Cn = self.Mtotal_wind[2] / q / s_ref / b_ref

        # Solves divide by zero error
        self.CL_over_CDi = cas.if_else(self.CDi == 0, 0, self.CL / self.CDi)

    def calculate_Vij(self,
                      points,  # type: cas.MX
                      align_trailing_vortices_with_freestream=True,  # Otherwise, aligns with x-axis
                      ):
        # Calculates Vij, the velocity influence matrix (First index is collocation point number, second index is vortex number).
        # points: the list of points (Nx3) to calculate the velocity influence at.

        n_points = points.shape[0]

        # Make a and b vectors.
        # a: Vector from all collocation points to all horseshoe vortex left vertices.
        #   # First index is collocation point #, second is vortex #.
        # b: Vector from all collocation points to all horseshoe vortex right vertices.
        #   # First index is collocation point #, second is vortex #.
        a_x = points[:, 0] - cas.repmat(cas.transpose(self.left_vortex_vertices[:, 0]), n_points, 1)
        a_y = points[:, 1] - cas.repmat(cas.transpose(self.left_vortex_vertices[:, 1]), n_points, 1)
        a_z = points[:, 2] - cas.repmat(cas.transpose(self.left_vortex_vertices[:, 2]), n_points, 1)
        b_x = points[:, 0] - cas.repmat(cas.transpose(self.right_vortex_vertices[:, 0]), n_points, 1)
        b_y = points[:, 1] - cas.repmat(cas.transpose(self.right_vortex_vertices[:, 1]), n_points, 1)
        b_z = points[:, 2] - cas.repmat(cas.transpose(self.right_vortex_vertices[:, 2]), n_points, 1)

        if align_trailing_vortices_with_freestream:
            freestream_direction = self.op_point.compute_freestream_direction_geometry_axes()
            u_x = freestream_direction[0]
            u_y = freestream_direction[1]
            u_z = freestream_direction[2]
        else:
            u_x = 1
            u_y = 0
            u_z = 0

        # Do some useful arithmetic
        a_cross_b_x = a_y * b_z - a_z * b_y
        a_cross_b_y = a_z * b_x - a_x * b_z
        a_cross_b_z = a_x * b_y - a_y * b_x
        a_dot_b = a_x * b_x + a_y * b_y + a_z * b_z

        a_cross_u_x = a_y * u_z - a_z * u_y
        a_cross_u_y = a_z * u_x - a_x * u_z
        a_cross_u_z = a_x * u_y - a_y * u_x
        a_dot_u = a_x * u_x + a_y * u_y + a_z * u_z

        b_cross_u_x = b_y * u_z - b_z * u_y
        b_cross_u_y = b_z * u_x - b_x * u_z
        b_cross_u_z = b_x * u_y - b_y * u_x
        b_dot_u = b_x * u_x + b_y * u_y + b_z * u_z

        norm_a = cas.sqrt(a_x ** 2 + a_y ** 2 + a_z ** 2)
        norm_b = cas.sqrt(b_x ** 2 + b_y ** 2 + b_z ** 2)
        norm_a_inv = 1 / norm_a
        norm_b_inv = 1 / norm_b

        # # Handle the special case where the collocation point is along a bound vortex leg
        # a_cross_b_squared = (
        #         a_cross_b_x ** 2 +
        #         a_cross_b_y ** 2 +
        #         a_cross_b_z ** 2
        # )
        # a_dot_b = cas.if_else(a_cross_b_squared < 1e-8, a_dot_b + 1, a_dot_b)
        # a_cross_u_squared = (
        #         a_cross_u_x ** 2 +
        #         a_cross_u_y ** 2 +
        #         a_cross_u_z ** 2
        # )
        # a_dot_u = cas.if_else(a_cross_u_squared < 1e-8, a_dot_u + 1, a_dot_u)
        # b_cross_u_squared = (
        #         b_cross_u_x ** 2 +
        #         b_cross_u_y ** 2 +
        #         b_cross_u_z ** 2
        # )
        # b_dot_u = cas.if_else(b_cross_u_squared < 1e-8, b_dot_u + 1, b_dot_u)

        # Handle the special case where the collocation point is along the bound vortex leg
        a_dot_b -= 1e-8
        # a_dot_xhat += 1e-8
        # b_dot_xhat += 1e-8

        # Calculate Vij
        term1 = (norm_a_inv + norm_b_inv) / (norm_a * norm_b + a_dot_b)
        term2 = norm_a_inv / (norm_a - a_dot_u)
        term3 = norm_b_inv / (norm_b - b_dot_u)

        Vij_x = 1 / (4 * np.pi) * (
                a_cross_b_x * term1 +
                a_cross_u_x * term2 -
                b_cross_u_x * term3
        )
        Vij_y = 1 / (4 * np.pi) * (
                a_cross_b_y * term1 +
                a_cross_u_y * term2 -
                b_cross_u_y * term3
        )
        Vij_z = 1 / (4 * np.pi) * (
                a_cross_b_z * term1 +
                a_cross_u_z * term2 -
                b_cross_u_z * term3
        )

        return Vij_x, Vij_y, Vij_z

    # def calculate_delta_cp(self):
    #     # Find the area of each panel ()
    #     diag1 = self.front_left_vertices - self.back_right_vertices
    #     diag2 = self.front_right_vertices - self.back_left_vertices
    #     self.areas = np.linalg.norm(np.cross(diag1, diag2, axis=1), axis=1) / 2
    #
    #     # Calculate panel data
    #     self.Fi_normal = np.einsum('ij,ij->i', self.forces_inviscid_geometry, self.normal_directions)
    #     self.pressure_normal = self.Fi_normal / self.areas
    #     self.delta_cp = self.pressure_normal / self.op_point.dynamic_pressure()

    def get_induced_velocity_at_point(self, point):
        if not self.opti.return_status() == 'Solve_Succeeded':
            print("WARNING: This method should only be used after a solution has been found!!!\n"
                  "Running anyway for debugging purposes - this is likely to not work.")

        Vij_x, Vij_y, Vij_z = self.calculate_Vij(point)

        # vortex_strengths = self.opti.debug.value(self.vortex_strengths)

        Vi_x = Vij_x @ self.vortex_strengths
        Vi_y = Vij_y @ self.vortex_strengths
        Vi_z = Vij_z @ self.vortex_strengths

        get = lambda x: self.opti.debug.value(x)
        Vi_x = get(Vi_x)
        Vi_y = get(Vi_y)
        Vi_z = get(Vi_z)

        Vi = np.vstack((Vi_x, Vi_y, Vi_z)).T

        return Vi

    def get_velocity_at_point(self, point):
        # Input: attrib_name Nx3 numpy array of points that you would like to know the velocities at.
        # Output: attrib_name Nx3 numpy array of the velocities at those points.

        Vi = self.get_induced_velocity_at_point(point)

        freestream = self.op_point.compute_freestream_velocity_geometry_axes()

        V = cas.transpose(cas.transpose(Vi) + freestream)
        return V

    def calculate_streamlines(self,
                              seed_points=None,  # will be auto-calculated if not specified
                              n_steps=100,  # minimum of 2
                              length=None  # will be auto-calculated if not specified
                              ):

        if length is None:
            length = self.airplane.c_ref * 5
        if seed_points is None:
            seed_points = (self.back_left_vertices + self.back_right_vertices) / 2

        # Resolution
        length_per_step = length / n_steps

        # Initialize
        streamlines = [seed_points]

        # Iterate
        for step_num in range(1, n_steps):
            update_amount = self.get_velocity_at_point(streamlines[-1])
            norm_update_amount = cas.sqrt(
                update_amount[:, 0] ** 2 + update_amount[:, 1] ** 2 + update_amount[:, 2] ** 2)
            update_amount = length_per_step * update_amount / norm_update_amount
            streamlines.append(streamlines[-1] + update_amount)

        self.streamlines = streamlines

    def draw(self,
             data_to_plot=None,
             data_name=None,
             show=True,
             draw_streamlines=True,
             recalculate_streamlines=False
             ):
        """
        Draws the solution. Note: Must be called on a SOLVED AeroProblem object.
        To solve an AeroProblem, use opti.solve(). To substitute a solved solution, use ap = ap.substitute_solution(sol).
        :return:
        """
        if self.verbose:
            print("Drawing...")

        if self.verbose and not self.opti.return_status() == 'Solve_Succeeded':
            print("WARNING: This method should only be used after a solution has been found!\n"
                  "Running anyway for debugging purposes - this is likely to not work...")

        # Do substitutions
        get = lambda x: self.opti.debug.value(x)
        front_left_vertices = get(self.front_left_vertices)
        front_right_vertices = get(self.front_right_vertices)
        back_left_vertices = get(self.back_left_vertices)
        back_right_vertices = get(self.back_right_vertices)
        left_vortex_vertices = get(self.left_vortex_vertices)
        right_vortex_vertices = get(self.right_vortex_vertices)
        self.vortex_strengths = get(self.vortex_strengths)
        try:
            data_to_plot = get(data_to_plot)
        except NotImplementedError:
            pass

        if data_to_plot is None:
            data_name = "Vortex Strengths"
            data_to_plot = get(self.vortex_strengths)

        fig = Figure3D()

        for index in range(len(front_left_vertices)):
            fig.add_quad(
                points=[
                    front_left_vertices[index, :],
                    front_right_vertices[index, :],
                    back_right_vertices[index, :],
                    back_left_vertices[index, :],
                ],
                intensity=data_to_plot[index],
                outline=True,
            )
            # fig.add_line( # Don't draw the quarter-chords
            #     points=[
            #         left_vortex_vertices[index],
            #         right_vortex_vertices[index]
            #     ],
            # )

        if draw_streamlines:
            if (not hasattr(self, 'streamlines')) or recalculate_streamlines:
                if self.verbose:
                    print("Calculating streamlines...")
                is_trailing_edge = np.array(self.is_trailing_edge, dtype=bool)
                seed_points = (back_left_vertices[is_trailing_edge] + back_right_vertices[is_trailing_edge]) / 2
                self.calculate_streamlines(seed_points=seed_points)

            if self.verbose:
                print("Parsing streamline data...")
            n_streamlines = self.streamlines[0].shape[0]
            n_timesteps = len(self.streamlines)

            for streamlines_num in range(n_streamlines):
                streamline = [self.streamlines[ts][streamlines_num, :] for ts in range(n_timesteps)]
                fig.add_streamline(
                    points=streamline,
                )

        return fig.draw(
            show=show,
            colorbar_title=data_name
        )



from ..geometry import *
from ..performance import *

class AeroProblem(AeroSandboxObject):
    def __init__(self,
                 airplane, # type: Airplane
                 op_point, # type: OperatingPoint
                 ):
        self.airplane = airplane
        self.op_point = op_point


def mass_hpa_wing(
        span,
        chord,
        vehicle_mass,
        n_ribs,  # You should optimize on this, there's a trade between rib weight and LE sheeting weight!
        n_wing_sections=1,  # defaults to a single-section wing (be careful: can you disassemble/transport this?)
        ultimate_load_factor=1.75,  # default taken from Daedalus design
        type="cantilevered",  # "cantilevered", "one-wire", "multi-wire"
        t_over_c=0.128,  # default from DAE11
        include_spar=True,
        # Should we include the mass of the spar? Useful if you want to do your own primary structure calculations.
):
    """
    Finds the mass of the wing structure of a human powered aircraft (HPA), following Juan Cruz's correlations in
    http://journals.sfu.ca/ts/index.php/ts/article/viewFile/760/718
    :param span: wing span [m]
    :param chord: wing mean chord [m]
    :param vehicle_mass: aircraft gross weight [kg]
    :param n_ribs: number of ribs in the wing
    :param n_wing_sections: number of wing sections or panels (for disassembly?)
    :param ultimate_load_factor: ultimate load factor [unitless]
    :param type: Type of bracing: "cantilevered", "one-wire", "multi-wire"
    :param t_over_c: wing airfoil thickness-to-chord ratio
    :param include_spar: Should we include the mass of the spar? Useful if you want to do your own primary structure calculations. [boolean]
    :return: Wing structure mass [kg]
    """
    ### Primary structure
    if include_spar:
        if type == "cantilevered":
            mass_primary_spar = (
                    (span * 1.17e-1 + span ** 2 * 1.10e-2) *
                    (1 + (ultimate_load_factor * vehicle_mass / 100 - 2) / 4)
            )
        elif type == "one-wire":
            mass_primary_spar = (
                    (span * 3.10e-2 + span ** 2 * 7.56e-3) *
                    (1 + (ultimate_load_factor * vehicle_mass / 100 - 2) / 4)
            )
        elif type == "multi-wire":
            mass_primary_spar = (
                    (span * 1.35e-1 + span ** 2 * 1.68e-3) *
                    (1 + (ultimate_load_factor * vehicle_mass / 100 - 2) / 4)
            )
        else:
            raise ValueError("Bad input for 'type'!")

        mass_primary = mass_primary_spar * (
                11382.3 / 9222.2)  # accounts for rear spar, struts, fittings, kevlar x-bracing, and wing-fuselage mounts
    else:
        mass_primary = 0

    ### Secondary structure
    ratio_of_rib_spacing_to_chord = (span / n_ribs) / chord
    n_end_ribs = 2 * n_wing_sections - 2
    area = span * chord

    # Rib mass
    W_wr = n_ribs * (chord ** 2 * t_over_c * 5.50e-2 + chord * 1.91e-3)

    # End rib mass
    W_wer = n_end_ribs * (chord ** 2 * t_over_c * 6.62e-1 + chord * 6.57e-3)

    # LE sheeting mass
    W_wLE = 0.456 * (span ** 2 * ratio_of_rib_spacing_to_chord ** (4 / 3) / span)

    # TE mass
    W_wTE = span * 2.77e-2

    # Covering
    W_wc = area * 3.08e-2

    mass_secondary = W_wr + W_wer + W_wLE + W_wTE + W_wc

    return mass_primary + mass_secondary


def mass_wing_spar(
        span,
        mass_supported,
        ultimate_load_factor=1.75,  # default taken from Daedalus design
        n_booms=1,
):
    """
    Finds the mass of the spar for a wing on a single- or multi-boom lightweight aircraft. Model originally designed for solar aircraft.
    Data was fit to the range 30 < wing_span < 90 [m] and 50 < supported_mass < 800 [kg], but validity should extend somewhat beyond that.
    Extremely accurate fits within this range; R^2 > 0.99 for all fits.
    Source: AeroSandbox\studies\MultiBoomSparMass
    Assumptions:
        * Rectangular lift distribution (close enough, slightly conservative w.r.t. elliptical)
        * Constraint that local wing dihedral/anhedral angle must not exceed 10 degrees anywhere.
        * If multi-boom, assumes static-aerostructurally-optimal placement of the outer booms.
    :param span: Wing span [m]
    :param mass_supported: Total mass of all fuselages + tails
    :param ultimate_load_factor: Design load factor. Default taken from Daedalus design.
    :param n_booms: Number of booms on the design. Can be 1, 2, or 3. Assumes optimal placement of the outer booms.
    :return:
    """
    if n_booms == 1:
        c = 0.0082116578817492
        m = 0.3380385531034899
        n = 1.6500210118367851
    elif n_booms == 2:
        c = 0.0038966832809997
        m = 0.3472862388655697
        n = 1.6222231509034950
    elif n_booms == 3:
        c = 0.0059175282654185
        m = 0.3807869156408368
        n = 1.3862788430962647
    else:
        raise ValueError("Bad value of n_booms!")

    return c * (mass_supported * ultimate_load_factor) ** m * span ** n


def mass_hpa_stabilizer(
        span,
        chord,
        dynamic_pressure_at_manuever_speed,
        n_ribs,  # You should optimize on this, there's a trade between rib weight and LE sheeting weight!
        t_over_c=0.128,  # default from DAE11
        include_spar=True,
        # Should we include the mass of the spar? Useful if you want to do your own primary structure calculations.
):
    """
    Finds the mass of a stabilizer structure of a human powered aircraft (HPA), following Juan Cruz's correlations in
    http://journals.sfu.ca/ts/index.php/ts/article/viewFile/760/718
    Note: apply this once to BOTH the rudder and elevator!!!
    :param span: stabilizer span [m]
    :param chord: stabilizer mean chord [m]
    :param dynamic_pressure_at_manuever_speed: dynamic pressure at maneuvering speed [Pa]
    :param n_ribs: number of ribs in the wing
    :param t_over_c: wing airfoil thickness-to-chord ratio
    :param include_spar: Should we include the mass of the spar? Useful if you want to do your own primary structure calculations. [boolean]
    :return: Stabilizer structure mass [kg]
    """
    ### Primary structure
    area = span * chord
    q = dynamic_pressure_at_manuever_speed
    if include_spar:
        W_tss = (
                (span * 4.15e-2 + span ** 2 * 3.91e-3) *
                (1 + ((q * area) / 78.5 - 1) / 2)
        )

        mass_primary = W_tss
    else:
        mass_primary = 0

    ### Secondary structure
    ratio_of_rib_spacing_to_chord = (span / n_ribs) / chord

    # Rib mass
    W_tsr = n_ribs * (chord ** 2 * t_over_c * 1.16e-1 + chord * 4.01e-3)

    # Leading edge sheeting
    W_tsLE = 0.174 * (area ** 2 * ratio_of_rib_spacing_to_chord ** (4 / 3) / span)

    # Covering
    W_tsc = area * 1.93e-2

    mass_secondary = W_tsr + W_tsLE + W_tsc

    ### Totaling
    correction_factor = ((537.8 / (537.8 - 23.7 - 15.1)) * (623.3 / (623.3 - 63.2 - 8.1))) ** 0.5
    # geometric mean of Daedalus elevator and rudder corrections from misc. weight

    return (mass_primary + mass_secondary) * correction_factor


def mass_hpa_tail_boom(
        length_tail_boom,
        dynamic_pressure_at_manuever_speed,
        mean_tail_surface_area,
):
    """
    Finds the mass of a tail boom structure of a human powered aircraft (HPA), following Juan Cruz's correlations in
    http://journals.sfu.ca/ts/index.php/ts/article/viewFile/760/718
    Assumes a tubular tail boom of high modules (E > 228 GPa) graphite/epoxy
    :param length_tail_boom: length of the tail boom [m]. Calculated as distance from the wing 1/4 chord to the furthest tail surface.
    :param dynamic_pressure_at_manuever_speed: dynamic pressure at maneuvering speed [Pa]
    :param mean_tail_surface_area: mean of the areas of the tail surfaces (elevator, rudder)
    :return: mass of the tail boom [m]
    """
    l = length_tail_boom
    q = dynamic_pressure_at_manuever_speed
    area = mean_tail_surface_area
    w_tb = (l * 1.14e-1 + l ** 2 * 1.96e-2) * (1 + ((q * area) / 78.5 - 1) / 2)

    return w_tb


def mass_surface_balsa_monokote_cf(
        chord,
        span,
        mean_t_over_c=0.08
):
    """
    Estimates the mass of a lifting surface constructed with balsa-monokote-carbon-fiber construction techniques.
    Warning: Not well validated; spar sizing is a guessed scaling and not based on structural analysis.
    :param chord: wing mean chord [m]
    :param span: wing span [m]
    :param mean_t_over_c: wing thickness-to-chord ratio [unitless]
    :return: estimated surface mass [kg]
    """
    mean_t = chord * mean_t_over_c
    ### Balsa wood + Monokote + a 1" dia CF tube spar.
    monokote_mass = 0.061 * chord * span * 2  # 0.2 oz/sqft

    rib_density = 200  # mass density, in kg/m^3
    rib_spacing = 0.1  # one rib every x meters
    rib_width = 0.003  # width of an individual rib
    ribs_mass = (
            (mean_t * chord * rib_width) *  # volume of a rib
            rib_density *  # density of a rib
            (span / rib_spacing)  # number of ribs
    )

    spar_mass_1_inch = 0.2113 * span * 1.5  # assuming 1.5x 1" CF tube spar
    spar_mass = spar_mass_1_inch * (
            mean_t / 0.0254) ** 2  # Rough GUESS for scaling, FIX THIS before using seriously!

    return (monokote_mass + ribs_mass + spar_mass) * 1.2  # for glue


def mass_surface_solid(
        chord,
        span,
        density=2700,  # kg/m^3, defaults to that of aluminum
        mean_t_over_c=0.08
):
    """
    Estimates the mass of a lifting surface constructed out of a solid piece of material.
    Warning: Not well validated; spar sizing is a guessed scaling and not based on structural analysis.
    :param chord: wing mean chord [m]
    :param span: wing span [m]
    :param mean_t_over_c: wing thickness-to-chord ratio [unitless]
    :return: estimated surface mass [kg]
    """
    mean_t = chord * mean_t_over_c
    volume = chord * span * mean_t
    return density * volume


if __name__ == "__main__":
    import casadi as cas
    import numpy as np
    import matplotlib.pyplot as plt

    # Daedalus wing mass validation
    print(
        "Daedalus wing, estimated mass: %f" %
        mass_hpa_wing(
            span=34,
            chord=0.902,
            vehicle_mass=104.1,
            n_ribs=100,
            n_wing_sections=5,
            type="one-wire"
        )
    )
    print(
        "Daedalus wing, actual mass: %f" % 18.9854
    )

    nr = np.linspace(1, 400, 401)
    m = mass_hpa_wing(
        span=34,
        chord=0.902,
        vehicle_mass=104.1,
        n_ribs=nr,
        n_wing_sections=5,
        type="one-wire"
    )
    plt.plot(nr, m)
    plt.ylim([15, 20])
    plt.grid(True)
    plt.xlabel("Number of ribs")
    plt.ylabel("Wing mass [kg]")
    plt.title("Daedalus Wing Rib Count Optimization Test")
    plt.show()

    # Test rib number optimization
    opti = cas.Opti()
    nr_opt = opti.variable()
    opti.set_initial(nr_opt, 100)
    opti.minimize(mass_hpa_wing(
        span=34,
        chord=0.902,
        vehicle_mass=104.1,
        n_ribs=nr_opt,
        n_wing_sections=5,
        type="one-wire"
    ))
    opti.solver('ipopt')
    sol = opti.solve()
    print("Optimal number of ribs: %f" % sol.value(nr_opt))

    print(
        "Daedalus elevator, estimated mass: %f" %
        mass_hpa_stabilizer(
            span=4.26,
            chord=0.6,
            dynamic_pressure_at_manuever_speed=1 / 2 * 1.225 * 7 ** 2,
            n_ribs=20,
        )
    )



### This file contains an assortment of random airfoils to use
from ..geometry import *
from .aerodynamics import *

generic_cambered_airfoil = Airfoil(
    CL_function=lambda alpha, Re, mach, deflection,: (  # Lift coefficient function
            (alpha * np.pi / 180) * (2 * np.pi) + 0.4550
    ),
    CDp_function=lambda alpha, Re, mach, deflection: (  # Profile drag coefficient function
            (1 + (alpha / 5) ** 2) * 2 * Cf_flat_plate(Re_L=Re)
    ),
    Cm_function=lambda alpha, Re, mach, deflection: (  # Moment coefficient function about quarter-chord
        -0.1
    )
)
generic_airfoil = Airfoil(
    CL_function=lambda alpha, Re, mach, deflection,: (  # Lift coefficient function
            (alpha * np.pi / 180) * (2 * np.pi)
    ),
    CDp_function=lambda alpha, Re, mach, deflection: (  # Profile drag coefficient function
            (1 + (alpha / 5) ** 2) * 2 * Cf_flat_plate(Re_L=Re)
    ),
    Cm_function=lambda alpha, Re, mach, deflection: (  # Moment coefficient function about quarter-chord
        0
    )  # TODO make this an actual curve!
)

# Make the airfoils
e216 = Airfoil("e216")
e216.CL_function = lambda alpha, Re, mach, deflection: (  # Lift coefficient function
    Cl_e216(alpha=alpha, Re_c=Re)
)
e216.CDp_function = lambda alpha, Re, mach, deflection: (  # Profile drag coefficient function
        Cd_profile_e216(alpha=alpha, Re_c=Re) +
        Cd_wave_e216(Cl=Cl_e216(alpha=alpha, Re_c=Re), mach=mach)
)
e216.Cm_function = lambda alpha, Re, mach, deflection: (  # Moment coefficient function about quarter-chord
    -0.15  # TODO
)  # TODO make this an actual curve!

rae2822 = Airfoil(
    CL_function=lambda alpha, Re, mach, deflection,: (  # Lift coefficient function
        Cl_rae2822(alpha=alpha, Re_c=Re)
    ),
    CDp_function=lambda alpha, Re, mach, deflection,: (  # Profile drag coefficient function
            Cd_profile_rae2822(alpha=alpha, Re_c=Re) +
            Cd_wave_rae2822(Cl=Cl_rae2822(alpha=alpha, Re_c=Re), mach=mach)
    ),
    Cm_function=lambda alpha, Re, mach, deflection: (  # Moment coefficient function about quarter-chord
        -0.05
    )  # TODO make this an actual curve!
)

flat_plate = Airfoil(
    CL_function=lambda alpha, Re, mach, deflection,: (  # Lift coefficient function
        Cl_flat_plate(alpha=alpha, Re_c=Re)
    ),
    CDp_function=lambda alpha, Re, mach, deflection,: (  # Profile drag coefficient function
            Cf_flat_plate(Re_L=Re) * 2
    ),
    Cm_function=lambda alpha, Re, mach, deflection: (  # Moment coefficient function
        0
    )  # TODO make this an actual curve!
)



from .aerodynamics import *

class panel1(AeroProblem):
    # 3D Panel Method legacy_aerodynamics code.
    #
    # Notes:
    #   # Specifically written to be reverse-mode-AD-compatible at every step.
    #
    # Usage:
    #   # Set up attrib_name problem using the syntax in the AeroProblem constructor (e.g. "panel1(airplane = attrib_name, op_point = op)" for some Airplane attrib_name and OperatingPoint op)
    #   # Call panel1.run() to run the problem.
    #   # Access results in the command line, or through properties of the panel1 class.
    #   #   # In attrib_name future update, this will be done through attrib_name standardized AeroData class.

    def run(self, verbose=True):
        self.verbose = verbose

        if self.verbose: print("Running PANEL1 calculation...")

        self.make_panels()
        self.setup_LHS()
        self.setup_RHS()
        self.calculate_vortex_strengths()
        # self.calculate_forces()

        if self.verbose: print("PANEL1 calculation complete!")

    def make_panels(self):
        # Creates self.panel_coordinates_structured_list and self.wing_mcl_normals.

        if self.verbose: print("Meshing...")

        collocation_points = np.empty((0, 3))
        normal_directions = np.empty((0, 3))
        # left_vortex_vertices = np.empty((0, 3))
        # right_vortex_vertices = np.empty((0, 3))
        front_left_vertices = np.empty((0, 3))
        front_right_vertices = np.empty((0, 3))
        back_left_vertices = np.empty((0, 3))
        back_right_vertices = np.empty((0, 3))
        areas = np.empty((0))
        is_trailing_edge_upper = np.empty((0), dtype=bool)
        is_trailing_edge_lower = np.empty((0), dtype=bool)

        for wing_num in range(len(self.airplane.wings)):

            # Get the wing
            wing = self.airplane.wings[wing_num]

            # Define number of chordwise points
            n_chordwise_coordinates = wing.chordwise_panels + 1

            # Get the chordwise coordinates
            if wing.chordwise_spacing == 'uniform':
                nondim_chordwise_coordinates = np.linspace(0, 1, n_chordwise_coordinates)
            elif wing.chordwise_spacing == 'cosine':
                nondim_chordwise_coordinates = cosspace(0, 1, n_chordwise_coordinates)
            else:
                raise Exception("Bad init_val of wing.chordwise_spacing!")

            # Get corners of xsecs
            xsec_xyz_le = np.empty((0, 3))  # Nx3 array of leading edge points
            xsec_xyz_te = np.empty((0, 3))  # Nx3 array of trailing edge points
            for xsec in wing.xsecs:
                xsec_xyz_le = np.vstack((xsec_xyz_le, xsec.xyz_le + wing.xyz_le))
                xsec_xyz_te = np.vstack((xsec_xyz_te, xsec.xyz_te() + wing.xyz_le))

            # Get quarter-chord vector
            xsec_xyz_quarter_chords = 0.75 * xsec_xyz_le + 0.25 * xsec_xyz_te  # Nx3 array of quarter-chord points
            section_quarter_chords = (
                    xsec_xyz_quarter_chords[1:, :] -
                    xsec_xyz_quarter_chords[:-1, :]
            )  # Nx3 array of vectors connecting quarter-chords

            # -----------------------------------------------------
            ## Get directions for transforming 2D airfoil data to 3D
            # First, project quarter chords onto YZ plane and normalize.
            section_quarter_chords_proj = (section_quarter_chords[:, 1:] /
                                           np.expand_dims(np.linalg.norm(section_quarter_chords[:, 1:], axis=1), axis=1)
                                           )  # Nx2 array of quarter-chord vectors projected onto YZ plane
            section_quarter_chords_proj = np.hstack(
                (np.zeros((section_quarter_chords_proj.shape[0], 1)), section_quarter_chords_proj)
            )  # Convert back to attrib_name Nx3 array, since that's what we'll need later.
            # Then, construct the normal directions for each xsec.
            if len(wing.xsecs) > 2:  # Make normals for the inner xsecs, where we need to merge directions
                xsec_local_normal_inners = section_quarter_chords_proj[:-1, :] + section_quarter_chords_proj[1:, :]
                xsec_local_normal_inners = (xsec_local_normal_inners /
                                            np.expand_dims(np.linalg.norm(xsec_local_normal_inners, axis=1), axis=1)
                                            )
                xsec_local_normal = np.vstack((
                    section_quarter_chords_proj[0, :],
                    xsec_local_normal_inners,
                    section_quarter_chords_proj[-1, :]
                ))
            else:
                xsec_local_normal = np.vstack((
                    section_quarter_chords_proj[0, :],
                    section_quarter_chords_proj[-1, :]
                ))
            # xsec_local_normal is now attrib_name Nx3 array that represents the normal direction at each xsec.
            # Then, construct the back directions for each xsec.
            xsec_local_back = xsec_xyz_te - xsec_xyz_le  # aligned with chord
            xsec_chord = np.linalg.norm(xsec_local_back, axis=1)  # 1D vector, one per xsec
            xsec_local_back = (xsec_local_back /
                               np.expand_dims(xsec_chord, axis=1)
                               )
            # Then, construct the up direction for each xsec.
            xsec_local_up = np.cross(xsec_local_back, xsec_local_normal,
                                     axis=1)  # Nx3 array that represents the upwards direction at each xsec.

            # Get the scaling factor (airfoils at dihedral breaks need to be "taller" to compensate)
            xsec_scaling_factor = 1 / np.sqrt((
                                                      1 + np.sum(
                                                  section_quarter_chords_proj[1:, :] * section_quarter_chords_proj[:-1,
                                                                                       :], axis=1
                                              )
                                              ) / 2
                                              )
            xsec_scaling_factor = np.hstack((1, xsec_scaling_factor, 1))  # TODO is this always right? idk

            # -----------------------------------------------------
            ## Make the panels for each section.

            for section_num in range(len(wing.xsecs) - 1):
                # Define the relevant cross sections
                inner_xsec = wing.xsecs[section_num]  # type: WingXSec
                outer_xsec = wing.xsecs[section_num + 1]  # type: WingXSec

                # Define the airfoils at each cross section
                inner_airfoil = inner_xsec.airfoil.get_sharp_TE_airfoil().get_airfoil_with_control_surface( #TODO consider whether this needs to be sharp
                    deflection=inner_xsec.control_surface_deflection,
                    hinge_point=inner_xsec.control_surface_hinge_point
                )  # type: Airfoil
                outer_airfoil = outer_xsec.airfoil.get_sharp_TE_airfoil().get_airfoil_with_control_surface(
                    deflection=inner_xsec.control_surface_deflection,
                    # inner xsec dictates control surface deflections.
                    hinge_point=inner_xsec.control_surface_hinge_point
                )  # type: Airfoil

                # Make the panel coordinates for each.
                inner_xsec_coordinates_nondim = inner_airfoil.get_repaneled_airfoil(
                    n_points_per_side=n_chordwise_coordinates
                ).coordinates
                outer_xsec_coordinates_nondim = outer_airfoil.get_repaneled_airfoil(
                    n_points_per_side=n_chordwise_coordinates
                ).coordinates

                # inner_xsec_coordinates: First index is point number, second index is xyz.
                inner_panel_coordinates = xsec_xyz_le[section_num, :] + (
                        xsec_local_back[section_num, :] * np.expand_dims(inner_xsec_coordinates_nondim[:, 0], 1) *
                        xsec_chord[
                            section_num] +
                        xsec_local_up[section_num, :] * np.expand_dims(inner_xsec_coordinates_nondim[:, 1], 1) *
                        xsec_chord[
                            section_num] * xsec_scaling_factor[section_num]
                )
                outer_panel_coordinates = xsec_xyz_le[section_num + 1, :] + (
                        xsec_local_back[section_num + 1, :] * np.expand_dims(outer_xsec_coordinates_nondim[:, 0], 1) *
                        xsec_chord[section_num + 1] +
                        xsec_local_up[section_num + 1, :] * np.expand_dims(outer_xsec_coordinates_nondim[:, 1], 1) *
                        xsec_chord[
                            section_num + 1] * xsec_scaling_factor[
                            section_num + 1]
                )

                # Define number of spanwise points
                n_spanwise_coordinates = xsec.spanwise_panels + 1

                # Get the spanwise coordinates
                if xsec.spanwise_spacing == 'uniform':
                    nondim_spanwise_coordinates = np.linspace(0, 1, n_spanwise_coordinates)
                elif xsec.spanwise_spacing == 'cosine':
                    nondim_spanwise_coordinates = cosspace(n_points=n_spanwise_coordinates)
                else:
                    raise Exception("Bad init_val of section.spanwise_spacing!")

                # Make section_panel_coordinates: MxNx3 array of the wing surface's coordinates.
                # First index is chordwise location, second index is spanwise location, third is xyz.
                section_panel_coordinates = (
                        np.expand_dims((1 - nondim_spanwise_coordinates), 2) * np.expand_dims(
                    inner_panel_coordinates, 1) +
                        np.expand_dims(nondim_spanwise_coordinates, 2) * np.expand_dims(
                    outer_panel_coordinates, 1)
                )  # TODO this is not strictly speaking correct, only true in the limit of small twist angles.

                # Compute corners of each panel
                front_inner_coordinates = section_panel_coordinates[1:, :-1, :]
                front_outer_coordinates = section_panel_coordinates[1:, 1:, :]
                back_inner_coordinates = section_panel_coordinates[:-1, :-1, :]
                back_outer_coordinates = section_panel_coordinates[:-1, 1:, :]
                section_is_trailing_edge_lower = np.vstack((
                    np.zeros((wing.chordwise_panels * 2 - 1, xsec.spanwise_panels), dtype=bool),
                    np.ones((1, xsec.spanwise_panels), dtype=bool)
                ))
                section_is_trailing_edge_upper = np.flipud(section_is_trailing_edge_lower)

                # Reshape
                front_inner_coordinates = np.reshape(front_inner_coordinates, (-1, 3), order='F')
                front_outer_coordinates = np.reshape(front_outer_coordinates, (-1, 3), order='F')
                back_inner_coordinates = np.reshape(back_inner_coordinates, (-1, 3), order='F')
                back_outer_coordinates = np.reshape(back_outer_coordinates, (-1, 3), order='F')
                section_is_trailing_edge_upper = np.reshape(section_is_trailing_edge_upper, (-1), order='F')
                section_is_trailing_edge_lower = np.reshape(section_is_trailing_edge_lower, (-1), order='F')

                # Calculate panel normals and areas via diagonals
                diag1 = front_outer_coordinates - back_inner_coordinates
                diag2 = front_inner_coordinates - back_outer_coordinates
                diag_cross = np.cross(diag1, diag2, axis=1)
                diag_cross_norm = np.linalg.norm(diag_cross, axis=1)
                normals_to_add = diag_cross / np.expand_dims(diag_cross_norm, axis=1)
                areas_to_add = diag_cross_norm / 2

                # Make the panel data
                collocations_to_add = (
                        0.5 * (0.5 * front_inner_coordinates + 0.5 * back_inner_coordinates) +
                        0.5 * (0.5 * front_outer_coordinates + 0.5 * back_outer_coordinates)
                )
                # inner_vortex_vertices_to_add = 0.75 * front_inner_coordinates + 0.25 * back_inner_coordinates
                # outer_vortex_vertices_to_add = 0.75 * front_outer_coordinates + 0.25 * back_outer_coordinates

                # Append to the lists of panel data (c, n, lv, rv, etc.)
                front_left_vertices = np.vstack((
                    front_left_vertices,
                    front_inner_coordinates
                ))
                front_right_vertices = np.vstack((
                    front_right_vertices,
                    front_outer_coordinates
                ))
                back_left_vertices = np.vstack((
                    back_left_vertices,
                    back_inner_coordinates
                ))
                back_right_vertices = np.vstack((
                    back_right_vertices,
                    back_outer_coordinates
                ))
                areas = np.hstack((
                    areas,
                    areas_to_add
                ))
                is_trailing_edge_upper = np.hstack((
                    is_trailing_edge_upper,
                    section_is_trailing_edge_upper
                ))
                is_trailing_edge_lower = np.hstack((
                    is_trailing_edge_lower,
                    section_is_trailing_edge_lower
                ))
                collocation_points = np.vstack((
                    collocation_points,
                    collocations_to_add
                ))
                normal_directions = np.vstack((
                    normal_directions,
                    normals_to_add
                ))
                # left_vortex_vertices = np.vstack((
                #     left_vortex_vertices,
                #     inner_vortex_vertices_to_add
                # ))
                # right_vortex_vertices = np.vstack((
                #     right_vortex_vertices,
                #     outer_vortex_vertices_to_add
                # ))

                # Handle symmetry
                if wing.symmetric:
                    if inner_xsec.control_surface_type == "asymmetric":
                        # # Define the relevant cross sections
                        # inner_xsec = wing.xsecs[section_num]  # type: WingXSec
                        # outer_xsec = wing.xsecs[section_num + 1]  # type: WingXSec

                        # Define the airfoils at each cross section
                        inner_airfoil = inner_xsec.airfoil.get_sharp_TE_airfoil().get_airfoil_with_control_surface(
                            deflection=-inner_xsec.control_surface_deflection,
                            hinge_point=inner_xsec.control_surface_hinge_point
                        )
                        outer_airfoil = outer_xsec.airfoil.get_sharp_TE_airfoil().get_airfoil_with_control_surface(
                            deflection=-inner_xsec.control_surface_deflection,
                            # inner xsec dictates control surface deflections.
                            hinge_point=inner_xsec.control_surface_hinge_point
                        )

                        # Make the panel coordinates for each.
                        inner_xsec_coordinates_nondim = inner_airfoil.get_repaneled_airfoil(
                            n_points_per_side=n_chordwise_coordinates
                        ).coordinates
                        outer_xsec_coordinates_nondim = outer_airfoil.get_repaneled_airfoil(
                            n_points_per_side=n_chordwise_coordinates
                        ).coordinates

                        # inner_xsec_mcl: First index is point number, second index is xyz.
                        inner_panel_coordinates = xsec_xyz_le[section_num, :] + (
                                xsec_local_back[section_num, :] * np.expand_dims(inner_xsec_coordinates_nondim[:, 0],
                                                                                 1) *
                                xsec_chord[
                                    section_num] +
                                xsec_local_up[section_num, :] * np.expand_dims(inner_xsec_coordinates_nondim[:, 1], 1) *
                                xsec_chord[
                                    section_num] * xsec_scaling_factor[section_num]
                        )
                        outer_panel_coordinates = xsec_xyz_le[section_num + 1, :] + (
                                xsec_local_back[section_num + 1, :] * np.expand_dims(
                            outer_xsec_coordinates_nondim[:, 0], 1) *
                                xsec_chord[section_num + 1] +
                                xsec_local_up[section_num + 1, :] * np.expand_dims(outer_xsec_coordinates_nondim[:, 1],
                                                                                   1) *
                                xsec_chord[
                                    section_num + 1] * xsec_scaling_factor[
                                    section_num + 1]
                        )
                        section_panel_coordinates = (
                                np.expand_dims((1 - nondim_spanwise_coordinates), 2) * np.expand_dims(
                            inner_panel_coordinates, 1) +
                                np.expand_dims(nondim_spanwise_coordinates, 2) * np.expand_dims(
                            outer_panel_coordinates, 1)
                        )  # TODO this is not strictly speaking correct, only true in the limit of small twist angles.

                        # Compute corners of each panel
                        front_inner_coordinates = section_panel_coordinates[1:, :-1, :]
                        front_outer_coordinates = section_panel_coordinates[1:, 1:, :]
                        back_inner_coordinates = section_panel_coordinates[:-1, :-1, :]
                        back_outer_coordinates = section_panel_coordinates[:-1, 1:, :]

                        # Reshape
                        front_inner_coordinates = np.reshape(front_inner_coordinates, (-1, 3), order='F')
                        front_outer_coordinates = np.reshape(front_outer_coordinates, (-1, 3), order='F')
                        back_inner_coordinates = np.reshape(back_inner_coordinates, (-1, 3), order='F')
                        back_outer_coordinates = np.reshape(back_outer_coordinates, (-1, 3), order='F')

                        # Calculate panel normals and areas via diagonals
                        diag1 = front_outer_coordinates - back_inner_coordinates
                        diag2 = front_inner_coordinates - back_outer_coordinates
                        diag_cross = np.cross(diag1, diag2, axis=1)
                        diag_cross_norm = np.linalg.norm(diag_cross, axis=1)
                        normals_to_add = diag_cross / np.expand_dims(diag_cross_norm, axis=1)
                        areas_to_add = diag_cross_norm / 2

                        # Make the panels and append them to the lists of panel data (c, n, lv, rv, etc.)
                        collocations_to_add = (
                                0.5 * (0.5 * front_inner_coordinates + 0.5 * back_inner_coordinates) +
                                0.5 * (0.5 * front_outer_coordinates + 0.5 * back_outer_coordinates)
                        )
                        # inner_vortex_vertices_to_add = 0.75 * front_inner_coordinates + 0.25 * back_inner_coordinates
                        # outer_vortex_vertices_to_add = 0.75 * front_outer_coordinates + 0.25 * back_outer_coordinates

                    front_left_vertices = np.vstack((
                        front_left_vertices,
                        reflect_over_XZ_plane(front_outer_coordinates)
                    ))
                    front_right_vertices = np.vstack((
                        front_right_vertices,
                        reflect_over_XZ_plane(front_inner_coordinates)
                    ))
                    back_left_vertices = np.vstack((
                        back_left_vertices,
                        reflect_over_XZ_plane(back_outer_coordinates)
                    ))
                    back_right_vertices = np.vstack((
                        back_right_vertices,
                        reflect_over_XZ_plane(back_inner_coordinates)
                    ))
                    areas = np.hstack((
                        areas,
                        areas_to_add
                    ))
                    is_trailing_edge_upper = np.hstack((
                        is_trailing_edge_upper,
                        section_is_trailing_edge_upper
                    ))
                    is_trailing_edge_lower = np.hstack((
                        is_trailing_edge_lower,
                        section_is_trailing_edge_lower
                    ))
                    collocation_points = np.vstack((
                        collocation_points,
                        reflect_over_XZ_plane(collocations_to_add)
                    ))
                    normal_directions = np.vstack((
                        normal_directions,
                        reflect_over_XZ_plane(normals_to_add)
                    ))
                    # left_vortex_vertices = np.vstack((
                    #     left_vortex_vertices,
                    #     reflect_over_XZ_plane(outer_vortex_vertices_to_add)
                    # ))
                    # right_vortex_vertices = np.vstack((
                    #     right_vortex_vertices,
                    #     reflect_over_XZ_plane(inner_vortex_vertices_to_add)
                    # ))

        # Write to self object
        self.front_left_vertices = front_left_vertices
        self.front_right_vertices = front_right_vertices
        self.back_left_vertices = back_left_vertices
        self.back_right_vertices = back_right_vertices
        self.areas = areas
        self.is_trailing_edge_upper = is_trailing_edge_upper
        self.is_trailing_edge_lower = is_trailing_edge_lower
        self.collocation_points = collocation_points
        self.normal_directions = normal_directions

        # Make the horseshoe vortex data. Trail off upper surface.
        self.left_horseshoe_vortex_vertices = back_left_vertices[is_trailing_edge_upper]
        self.right_horseshoe_vortex_vertices = back_right_vertices[is_trailing_edge_upper]

        self.horseshoe_vortex_centers = (self.left_horseshoe_vortex_vertices + self.right_horseshoe_vortex_vertices) / 2

        # Do final processing for later use
        # self.vortex_centers = (self.left_vortex_vertices + self.right_vortex_vertices) / 2
        # self.vortex_bound_leg = (self.right_vortex_vertices - self.left_vortex_vertices)
        self.n_panels = len(self.collocation_points)
        self.n_horseshoes = len(self.left_horseshoe_vortex_vertices)

        if self.verbose: print("Meshing complete!")
        # -----------------------------------------------------

    def setup_LHS(self):
        # Sets up the left hand side of the governing equations.
        # This is governed by the geometry.
        # The left hand side consists of attrib_name matrix as follows:
        #   | *                  * *         * |
        #   |                                  |
        #   |      AIC (Vij*n)         AIC     |
        #   |     for doublets         for     |
        #   |                       horseshoes |
        #   |                                  |
        #   | *                  * *         * |
        #   | *                  * *         * |
        #   |   Kutta Condition      Zeroes    |
        #   | *    submatrix     * *         * |

        # # Calculate AIC matrix
        # ----------------------
        if self.verbose: print("Calculating the collocation influence matrix...")
        self.Vij = self.calculate_Vij(self.collocation_points)

        # AIC = (Vij * normal vectors)
        self.AIC = np.sum(
            self.Vij * np.expand_dims(self.normal_directions, 1),
            axis=2
        ) #

        # Kutta Conditions
        if self.verbose: print("Enforcing Kutta condition...")
        trailing_edge_upper_indices = np.argwhere(self.is_trailing_edge_upper)
        trailing_edge_lower_indices = np.argwhere(self.is_trailing_edge_lower)

        kutta_condition_submatrix = np.zeros((self.n_horseshoes, self.n_panels))
        np.put_along_axis(
            kutta_condition_submatrix,
            trailing_edge_upper_indices,
            1,
            axis=1
        )
        np.put_along_axis(
            kutta_condition_submatrix,
            trailing_edge_lower_indices,
            -1,
            axis=1
        )

        # Make the zeros submatrix
        zeros_submatrix = -np.eye(self.n_horseshoes) #np.zeros((self.n_horseshoes, self.n_horseshoes))

        # Merge everything
        kutta_and_zeros = np.hstack((kutta_condition_submatrix, zeros_submatrix))
        self.LHS_matrix = np.vstack((self.AIC, kutta_and_zeros))


        # # Calculate Vij at vortex centers for force calculation
        # -------------------------------------------------------
        # if self.verbose: print("Calculating the vortex center influence matrix...")
        # self.Vij_centers = self.calculate_Vij(self.vortex_centers)

        # # LU Decomposition on AIC
        # -------------------------
        # Unfortunately, I don't think we can use sp_linalg.lu_factor with autograd, so we'll have to do attrib_name direct solve
        # for every op-point instead of saving an LU-factorization and reusing it.
        # This isn't the worst, though, since the solution time is very small compared to the AIC calculation time, and
        # autograd gives us good gradients to use for op-point trimming and other things. So that's nice. I guess.
        # The long-term solution here would be to write attrib_name vector-jacobian product in autograd for lu_factor and lu_solve.

        # Old code
        # if self.verbose: print("LU factorizing the AIC matrix...")
        # self.lu, self.piv = sp_linalg.lu_factor(self.AIC) # TODO consider whether lu_factor is possible w autograd

    def setup_RHS(self):  # TODO hasn't been checked yet
        # Sets up the right hand side of the governing equations
        # This is dominated by the operating conditions.

        if self.verbose: print("Calculating the freestream influence...")
        self.steady_freestream_velocity = np.expand_dims(self.op_point.compute_freestream_velocity_geometry_axes(),
                                                         0)  # Direction the wind is GOING TO, in geometry axes coordinates
        self.rotation_freestream_velocities = self.op_point.compute_rotation_velocity_geometry_axes(
            self.collocation_points)

        self.freestream_velocities = self.steady_freestream_velocity + self.rotation_freestream_velocities  # Nx3, represents the freestream velocity at each panel collocation point (c)

        self.freestream_influences = np.sum(self.freestream_velocities * self.normal_directions, axis=1)

        # Kutta condition set to zero
        kutta_zeros = np.zeros(self.n_horseshoes)

        # Sets up the RHS
        self.RHS_vector = np.hstack((-self.freestream_influences, kutta_zeros))

    def calculate_vortex_strengths(self):
        # # Calculate Vortex Strengths
        # ----------------------------
        # Governing Equation: AIC @ Gamma + freestream_influence = 0
        if self.verbose: print("Calculating vortex strengths...")
        self.solution = np.linalg.solve(self.LHS_matrix, self.RHS_vector)

        self.doublet_strengths = self.solution[:self.n_panels]
        self.horseshoe_vortex_strengths = self.solution[self.n_panels:]
        print("horseshoe_vortex_strengths sum: ", np.sum(self.horseshoe_vortex_strengths))

    def calculate_forces(self): # TODO rewrite this for the new velocity influence matrix
        # # Calculate Near-Field Forces and Moments
        # -----------------------------------------
        # Governing Equation: The force on attrib_name straight, small vortex filament is F = rho * V p l * gamma,
        # where rho is density, V is the velocity vector, p is the cross product operator,
        # l is the vector of the filament itself, and gamma is the circulation.

        if self.verbose: print("Calculating forces on each panel...")
        # Calculate Vi (local velocity at the ith vortex center point)
        Vi_x = self.Vij_centers[:, :, 0] @ self.vortex_strengths + self.freestream_velocities[:, 0]
        Vi_y = self.Vij_centers[:, :, 1] @ self.vortex_strengths + self.freestream_velocities[:, 1]
        Vi_z = self.Vij_centers[:, :, 2] @ self.vortex_strengths + self.freestream_velocities[:, 2]
        Vi_x = np.expand_dims(Vi_x, axis=1)
        Vi_y = np.expand_dims(Vi_y, axis=1)
        Vi_z = np.expand_dims(Vi_z, axis=1)
        Vi = np.hstack((Vi_x, Vi_y, Vi_z))

        # Calculate forces_inviscid_geometry, the force on the ith panel. Note that this is in GEOMETRY AXES,
        # not WIND AXES or BODY AXES.
        density = self.op_point.density
        Vi_cross_li = np.cross(Vi, self.vortex_bound_leg, axis=1)
        vortex_strengths_expanded = np.expand_dims(self.vortex_strengths, axis=1)
        self.Fi_geometry = density * Vi_cross_li * vortex_strengths_expanded

        # Calculate total forces and moments
        if self.verbose: print("Calculating total forces and moments...")
        self.Ftotal_geometry = np.sum(self.Fi_geometry,
                                      axis=0)  # Remember, this is in GEOMETRY AXES, not WIND AXES or BODY AXES.
        # if self.verbose: print("Total aerodynamic forces (geometry axes): ", self.force_total_inviscid_geometry)

        self.Ftotal_wind = np.transpose(self.op_point.compute_rotation_matrix_wind_to_geometry()) @ self.Ftotal_geometry
        # if self.verbose: print("Total aerodynamic forces (wind axes):", self.force_total_inviscid_wind)

        self.Mtotal_geometry = np.sum(np.cross(self.vortex_centers - self.airplane.xyz_ref, self.Fi_geometry),
                                      axis=0)
        self.Mtotal_wind = np.transpose(self.op_point.compute_rotation_matrix_wind_to_geometry()) @ self.Mtotal_geometry

        # Calculate nondimensional forces
        q = self.op_point.dynamic_pressure()
        s_ref = self.airplane.s_ref
        b_ref = self.airplane.b_ref
        c_ref = self.airplane.c_ref
        self.CL = -self.Ftotal_wind[2] / q / s_ref
        self.CDi = -self.Ftotal_wind[0] / q / s_ref
        self.CY = self.Ftotal_wind[1] / q / s_ref
        self.Cl = self.Mtotal_wind[0] / q / b_ref
        self.Cm = self.Mtotal_wind[1] / q / c_ref
        self.Cn = self.Mtotal_wind[2] / q / b_ref

        # Calculate nondimensional moments

        # Solves divide by zero error
        if self.CDi == 0:
            self.CL_over_CDi = 0
        else:
            self.CL_over_CDi = self.CL / self.CDi

        if self.verbose: print("\nForces\n-----")
        if self.verbose: print("CL: ", self.CL)
        if self.verbose: print("CDi: ", self.CDi)
        if self.verbose: print("CY: ", self.CY)
        if self.verbose: print("CL/CDi: ", self.CL_over_CDi)
        if self.verbose: print("\nMoments\n-----")
        if self.verbose: print("Cl: ", self.Cl)
        if self.verbose: print("Cm: ", self.Cm)
        if self.verbose: print("Cn: ", self.Cn)

    def calculate_Vij(self, points):
        Vij_doublets = self.calculate_Vij_doublets(points) # Calculates the section of Vij corresponding to the doublets (ring vortices)
        Vij_horseshoes = self.calculate_Vij_horseshoes(points) # Calculates the section of Vij corresponding to the trailing horseshoe vortices
        Vij = np.hstack((Vij_doublets, Vij_horseshoes))

        return Vij

    def calculate_Vij_doublets(self, points):
        # Calculates the doublet part of Vij, the velocity influence matrix (First index is collocation point number, second index is vortex number).
        # points: the list of points (Nx3) to calculate the velocity influence at.

        # Data cleanup
        points = np.reshape(points, (-1, 3))

        # Getting info
        n_points = len(points)
        n_vortices = self.n_panels

        # Make v1, v2, v3, and v4 vectors.
        # Each vector goes from all collocation points to one type of vertex (front left, front right, etc.). NxNx3.
        #   # First index is collocation point #, second is vortex #, and third is xyz. N=n_panels
        # v1: corresponds to front left vertices
        # v2: corresponds to front right vertices
        # v3: corresponds to back right vertices
        # v4: corresponds to back left vertices
        # Example: v1[i,j,:] = collocation_points[i,:] - front_left_vertices[j,:]

        points = np.expand_dims(points, 1)
        v1 = points - self.front_left_vertices
        v2 = points - self.front_right_vertices
        v3 = points - self.back_right_vertices
        v4 = points - self.back_left_vertices

        # x_hat = np.zeros([n_points, n_vortices, 3])
        # x_hat[:, :, 0] = 1

        # Do some useful arithmetic
        v1_cross_v2 = np.cross(v1, v2, axis = 2)
        v2_cross_v3 = np.cross(v2, v3, axis = 2)
        v3_cross_v4 = np.cross(v3, v4, axis = 2)
        v4_cross_v1 = np.cross(v4, v1, axis = 2)
        v1_dot_v2 = np.einsum('ijk,ijk->ij', v1, v2)
        v2_dot_v3 = np.einsum('ijk,ijk->ij', v2, v3)
        v3_dot_v4 = np.einsum('ijk,ijk->ij', v3, v4)
        v4_dot_v1 = np.einsum('ijk,ijk->ij', v4, v1)
        norm_v1 = np.linalg.norm(v1, axis = 2)
        norm_v2 = np.linalg.norm(v2, axis = 2)
        norm_v3 = np.linalg.norm(v3, axis = 2)
        norm_v4 = np.linalg.norm(v4, axis = 2)
        norm_v1_inv = 1 / norm_v1
        norm_v2_inv = 1 / norm_v2
        norm_v3_inv = 1 / norm_v3
        norm_v4_inv = 1 / norm_v4


        # Check for the special case where the collocation point is along the bound vortex leg
        # Find where cross product is near zero, and set the dot product to infinity so that the init_val of the bound term is zero.
        v1_v2_singularity_indices = (
                np.einsum('ijk,ijk->ij', v1_cross_v2, v1_cross_v2)  # norm(cross_product)^2
                < 3.0e-16)
        v1_dot_v2 = v1_dot_v2 + v1_v2_singularity_indices

        v2_v3_singularity_indices = (
                np.einsum('ijk,ijk->ij', v2_cross_v3, v2_cross_v3)  # norm(cross_product)^2
                < 3.0e-16)
        v2_dot_v3 = v2_dot_v3 + v2_v3_singularity_indices

        v3_v4_singularity_indices = (
                np.einsum('ijk,ijk->ij', v3_cross_v4, v3_cross_v4)  # norm(cross_product)^2
                < 3.0e-16)
        v3_dot_v4 = v3_dot_v4 + v3_v4_singularity_indices

        v4_v1_singularity_indices = (
                np.einsum('ijk,ijk->ij', v4_cross_v1, v4_cross_v1)  # norm(cross_product)^2
                < 3.0e-16)
        v4_dot_v1 = v4_dot_v1 + v4_v1_singularity_indices


        # Calculate Vij
        term1 = (norm_v1_inv + norm_v2_inv) / (norm_v1 * norm_v2 + v1_dot_v2)
        term1 = np.expand_dims(term1, 2)
        term2 = (norm_v2_inv + norm_v3_inv) / (norm_v2 * norm_v3 + v2_dot_v3)
        term2 = np.expand_dims(term2, 2)
        term3 = (norm_v3_inv + norm_v4_inv) / (norm_v3 * norm_v4 + v3_dot_v4)
        term3 = np.expand_dims(term3, 2)
        term4 = (norm_v4_inv + norm_v1_inv) / (norm_v4 * norm_v1 + v4_dot_v1)
        term4 = np.expand_dims(term4, 2)


        Vij_doublets = 1 / (4 * np.pi) * (
                v1_cross_v2 * term1 +
                v2_cross_v3 * term2 +
                v3_cross_v4 * term3 +
                v4_cross_v1 * term4
        )

        return Vij_doublets

    def calculate_Vij_horseshoes(self,points):
        # Calculates Vij, the velocity influence matrix (First index is collocation point number, second index is vortex number).
        # points: the list of points (Nx3) to calculate the velocity influence at.

        # Make lv and rv
        left_vortex_vertices = self.left_horseshoe_vortex_vertices
        right_vortex_vertices = self.right_horseshoe_vortex_vertices

        points = np.reshape(points, (-1, 3))
        n_points = len(points)
        n_vortices = len(left_vortex_vertices)

        # Make attrib_name and b vectors.
        # attrib_name: Vector from all collocation points to all horseshoe vortex left  vertices, NxNx3.
        #   # First index is collocation point #, second is vortex #, and third is xyz. N=n_panels
        # b: Vector from all collocation points to all horseshoe vortex right vertices, NxNx3.
        #   # First index is collocation point #, second is vortex #, and third is xyz. N=n_panels
        # attrib_name[i,j,:] = c[i,:] - lv[j,:]
        # b[i,j,:] = c[i,:] - rv[j,:]
        points = np.expand_dims(points, 1)
        a = points - left_vortex_vertices
        b = points - right_vortex_vertices
        # x_hat = np.zeros([n_points, n_vortices, 3])
        # x_hat[:, :, 0] = 1

        # Do some useful arithmetic
        a_cross_b = np.cross(a, b, axis=2)
        a_dot_b = np.einsum('ijk,ijk->ij', a, b)

        a_cross_x = np.stack((
            np.zeros((n_points, n_vortices)),
            a[:, :, 2],
            -a[:, :, 1]
        ), axis=2)
        a_dot_x = a[:, :, 0]

        b_cross_x = np.stack((
            np.zeros((n_points, n_vortices)),
            b[:, :, 2],
            -b[:, :, 1]
        ), axis=2)
        b_dot_x = b[:, :, 0]  # np.sum(b * x_hat,axis=2)

        norm_a = np.linalg.norm(a, axis=2)
        norm_b = np.linalg.norm(b, axis=2)
        norm_a_inv = 1 / norm_a
        norm_b_inv = 1 / norm_b

        # Check for the special case where the collocation point is along the bound vortex leg
        # Find where cross product is near zero, and set the dot product to infinity so that the init_val of the bound term is zero.
        bound_vortex_singularity_indices = (
                np.einsum('ijk,ijk->ij', a_cross_b, a_cross_b)  # norm(a_cross_b) ** 2
                < 3.0e-16)
        a_dot_b = a_dot_b + bound_vortex_singularity_indices
        left_vortex_singularity_indices = (
                np.einsum('ijk,ijk->ij', a_cross_x, a_cross_x)
                < 3.0e-16
        )
        a_dot_x = a_dot_x + left_vortex_singularity_indices
        right_vortex_singularity_indices = (
                np.einsum('ijk,ijk->ij', b_cross_x, b_cross_x)
                < 3.0e-16
        )
        b_dot_x = b_dot_x + right_vortex_singularity_indices

        # Calculate Vij
        term1 = (norm_a_inv + norm_b_inv) / (norm_a * norm_b + a_dot_b)
        term2 = (norm_a_inv) / (norm_a - a_dot_x)
        term3 = (norm_b_inv) / (norm_b - b_dot_x)
        term1 = np.expand_dims(term1, 2)
        term2 = np.expand_dims(term2, 2)
        term3 = np.expand_dims(term3, 2)

        Vij_horseshoes = 1 / (4 * np.pi) * (
                a_cross_b * term1 +
                a_cross_x * term2 -
                b_cross_x * term3
        )

        return Vij_horseshoes # TODO do this

    def calculate_delta_cp(self):
        # Find the area of each panel ()
        diag1 = self.front_left_vertices - self.back_right_vertices
        diag2 = self.front_right_vertices - self.back_left_vertices
        self.areas = np.linalg.norm(np.cross(diag1, diag2, axis=1), axis=1) / 2

        # Calculate panel data
        self.Fi_normal = np.einsum('ij,ij->i', self.Fi_geometry, self.normal_directions)
        self.pressure_normal = self.Fi_normal / self.areas
        self.delta_cp = self.pressure_normal / self.op_point.dynamic_pressure()

    def get_induced_velocity_at_point(self, point):
        # Input: attrib_name Nx3 numpy array of points that you would like to know the induced velocities at.
        # Output: attrib_name Nx3 numpy array of the induced velocities at those points.
        point = np.reshape(point, (-1, 3))

        Vij = self.calculate_Vij(point)

        solution_expanded = np.expand_dims(self.solution, 1)

        # freestream = self.op_point.compute_freestream_velocity_geometry_axes()
        # V_x = Vij[:, :, 0] @ vortex_strengths_expanded + freestream[0]
        # V_y = Vij[:, :, 1] @ vortex_strengths_expanded + freestream[1]
        # V_z = Vij[:, :, 2] @ vortex_strengths_expanded + freestream[2]

        Vi_x = Vij[:, :, 0] @ solution_expanded
        Vi_y = Vij[:, :, 1] @ solution_expanded
        Vi_z = Vij[:, :, 2] @ solution_expanded

        Vi = np.hstack((Vi_x, Vi_y, Vi_z))

        return Vi

    def get_velocity_at_point(self, point):
        # Input: attrib_name Nx3 numpy array of points that you would like to know the velocities at.
        # Output: attrib_name Nx3 numpy array of the velocities at those points.
        point = np.reshape(point, (-1, 3))

        Vi = self.get_induced_velocity_at_point(point)

        freestream = self.op_point.compute_freestream_velocity_geometry_axes()

        V = Vi + freestream
        return V

    def get_streamlines(self,
                        seed_points,  # A Nx3 ndarray of points to use for streamlines
                        n_steps = 300,  # number of points in streamline
                        length = 0.7,  # meters
                        ):
        # Calculates streamlines eminating from an array of given initial points.
        # "streamlines" is attrib_name MxNx3 array, where M is the index of the streamline number,
        # N is the index of the timestep, and the last index is xyz

        # Resolution
        length_per_step = length / n_steps
        n_streamlines = len(seed_points)

        # Initialize
        streamlines = np.zeros((n_streamlines, n_steps, 3))
        streamlines[:, 0, :] = seed_points

        # Iterate with forward Euler # TODO consider higher order methods like RK4
        for step_num in range(1, n_steps):
            update_amount = self.get_velocity_at_point(streamlines[:, step_num - 1, :])
            update_amount = update_amount * length_per_step / np.expand_dims(np.linalg.norm(update_amount, axis=1),
                                                                             axis=1)
            streamlines[:, step_num, :] = streamlines[:, step_num - 1, :] + update_amount

        return streamlines

    def draw(self,
             shading_type="solid",  # Can be None, "solid", "doublet_strengths", "delta_cp", or "trailing_edges"
             streamlines_type="trailing",  # Can be None, "trailing", "line", or attrib_name Nx3 numpy array with custom points.
             points_type = None, # You can supply the name of attrib_name property of this class to plot attrib_name point cloud. Useful for debugging, mostly.
             ):
        # Note: NOT autograd-compatible!
        print("Drawing...")
        plotter = pv.Plotter() # Initialize Plotter

        # Shading
        if shading_type is not None:

            # Make airplane geometry
            vertices = np.vstack((
                self.front_left_vertices,
                self.front_right_vertices,
                self.back_right_vertices,
                self.back_left_vertices
            ))
            faces = np.transpose(np.vstack((
                4 * np.ones(self.n_panels),
                np.arange(self.n_panels),
                np.arange(self.n_panels) + self.n_panels,
                np.arange(self.n_panels) + 2 * self.n_panels,
                np.arange(self.n_panels) + 3 * self.n_panels,
            )))
            faces = np.reshape(faces, (-1), order='C')
            wing_surfaces = pv.PolyData(vertices, faces)

            if shading_type == "solid":
                plotter.add_mesh(wing_surfaces, color='tan', show_edges=True,
                                 smooth_shading=True)
            elif shading_type == "doublet_strengths":
                if not hasattr(self, 'doublet_strengths'):
                    print("Doublet strengths not found, running again.")
                    self.run()

                # min = -1.5
                # max = 1.5
                scalars = self.doublet_strengths #np.minimum(np.maximum(self.delta_cp, delta_cp_min), delta_cp_max)

                cmap = plt.cm.get_cmap('viridis')
                plotter.add_mesh(wing_surfaces, scalars=scalars, cmap=cmap, color='tan', show_edges=True,
                                 smooth_shading=True)
                plotter.add_scalar_bar(title="Doublet Strengths", n_labels=5, shadow=True,
                                       font_family='arial')
            elif shading_type == "delta_cp":
                if not hasattr(self, 'delta_cp'):
                    self.calculate_delta_cp()

                delta_cp_min = -1.5
                delta_cp_max = 1.5

                scalars = np.minimum(np.maximum(self.delta_cp, delta_cp_min), delta_cp_max)
                cmap = plt.cm.get_cmap('viridis')
                plotter.add_mesh(wing_surfaces, scalars=scalars, cmap=cmap, color='tan', show_edges=True,
                                 smooth_shading=True)
                plotter.add_scalar_bar(title="Pressure Coefficient Differential", n_labels=5, shadow=True,
                                       font_family='arial')
            elif shading_type == "all_trailing_edges":
                scalars = np.logical_or(self.is_trailing_edge_upper, self.is_trailing_edge_lower) # type: np.ndarray
                scalars = scalars.astype(int)
                plotter.add_mesh(wing_surfaces, scalars=scalars, color='tan', show_edges=True,
                                 smooth_shading=True)
            elif shading_type == "upper_trailing_edges":
                scalars = self.is_trailing_edge_upper  # type: np.ndarray
                scalars = scalars.astype(int)
                plotter.add_mesh(wing_surfaces, scalars=scalars, color='tan', show_edges=True,
                                 smooth_shading=True)

        # Streamlines
        if streamlines_type is not None:
            if streamlines_type == "trailing":
                seed_points = np.vstack((
                    (0.5 * (self.back_left_vertices + self.back_right_vertices))[self.is_trailing_edge_upper],
                    (0.5 * (self.front_left_vertices + self.front_right_vertices))[self.is_trailing_edge_lower]
                ))
            elif streamlines_type == "line":
                seed_points = linspace_3D((0,0,-0.05),(0,0,0.05),30)
            else:
                seed_points = streamlines_type # assume streamlines_type is attrib_name numpy ndarray

            streamlines = self.get_streamlines(seed_points = seed_points)

            for streamline_num in range(len(streamlines)):
                plotter.add_lines(streamlines[streamline_num, :, :], width=1, color='#50C7C7')

        # Points
        if points_type is not None:
            points = getattr(self,points_type)

            plotter.add_points(points)


        # Do the plotting
        plotter.show_grid(color='#444444')
        plotter.set_background(color="black")
        plotter.show(cpos=(-1, -1, 1), full_screen=False)
